/**
 * This class represents a single AudioFile, regardless of whether it has been loaded or not.
 * @param {AudioContext} context The audio context.
 * @param {string} url The URL from which to load the file.
 * @param {GainNode} masterVolumeGain The master volume gain node, connected to the audio output.
 * @param {GainNode} effectsVolumeGain The effects volume gain node, connected to the master gain node.
 * @param {GainNode} backgroundVolumeGain The background volume gain node, connected to the master gain node.
 * @param {int} defaultVolume The default volume for the audio file, from 1 to 100.
 */
function AudioFile(context, url, masterVolumeGain, effectsVolumeGain, backgroundVolumeGain, defaultVolume) {
    // PRIVATE MEMBERS \\       var private = "test";
    var that = this;

    // Default playing volume.
    var ownVolume = defaultVolume || 50;
    ownVolume = ownVolume / 100;

    // Whether or not the AudioFile is currently loading.
    var loading = false;
    // Whether or not the AudioFile has successfully loaded and decoded.
    var loaded = false;
    // Whether or not there was an error during the loading of the AudioFile.
    var error = false;

    // Promise that will resolve when loading of the file succeeds.
    var loadPromise;
    // Resolve function for the promise.
    var loadPromiseResolve;
    // Reject function for the promise.
    var loadPromiseReject;

    // The audio buffer formed by decoding the loaded file.
    var audioBuffer;

    // Whether or not the looping audio is playing.
    var looping = false;

    // Whether or not the sound is queued to play a single time when loaded.
    var queuedPlay = false;

    // Whether or not a loop is queued to start when th
    var queuedLoop = false;

    /**
     * An object of all currently playing (not looping) nodes mapped to their id.
     */
    var currentlyPlayingNodes = {};
    var currentlyPlayingMaxId = 0;

    // The current loop node, which can be accessed to stop it.
    var loopNode;

    // The loop node's additional gain, which can be used to fade out the loop.
    var loopGain = context.createGain();
    loopGain.connect(backgroundVolumeGain);

    // PRIVATE FUNCTIONS \\     function privateFunction(){}

    /**
     * Function called when the audio data has finished decoding.
     * @param {binary} buffer The decoded buffer formed from the audio file.
     */
    function audioDataDecoded(buffer) {
        audioBuffer = buffer;
        loaded = true;
        loading = false;
        loadPromiseResolve();
    }

    /**
     * Called when the audio file is loaded.
     * "this" refers to the XMLHttpRequest used to load the file.
     */
    function fileLoaded() {
        context.decodeAudioData(this.response, audioDataDecoded, loadError);
    }

    /**
     * Function called when audio has failed to load or decode.
     */
    function loadError() {
        error = true;
        loading = false;
        loadPromiseReject();
    }

    /**
     * Creates the promise used to queue loops.
     */
    function createPromise() {
        loadPromise = new Promise(function (resolve, reject) {
            loadPromiseResolve = resolve;
            loadPromiseReject = reject;
        });
        loadPromise.then(processQueues);
    }

    /**
     * Creates and starts an XMLHttpRequest to fetch the audio data.
     */
    function startRequest() {
        var request = new XMLHttpRequest();
        request.open('GET', url);
        request.responseType = 'arraybuffer';
        request.onload = fileLoaded;
        request.onerror = loadError;
        request.send();
    }

    /**
     * Processes both the play and the loop load queue.
     */
    function processQueues() {
        if (queuedPlay) {
            that.play();
        }
        if (queuedLoop) {
            queuedLoop = false;
            startLoop();
        }
    }

    /**
     * Creates a volume gain set to the specified volume.
     * @param {int} volume Volume ranging from 0-1.
     * @returns {GainNode} The gain. Unconnected.
     */
    function createVolumeGain(volume) {
        var gain = context.createGain();
        gain.gain.setValueAtTime(volume, context.currentTime);
        return gain;
    }

    /**
     * Adds an object containing the node and the gain to the object of all currently playing nodes.
     * Also registers the onend function that will remove the node from the object when it finishes.
     * @param {SourceNode} node The node to add.
     * @param {GainNode} gain The gain to add.
     * @returns {int} The id the node was assigned.
     */
    function addToPlayingNodes(node, gain) {
        var id = currentlyPlayingMaxId;
        var object = {node: node, gain: gain};
        currentlyPlayingNodes[id] = object;
        node.onended = function (event) {
            removeFromPlayingNodes(id);
        };
        ++currentlyPlayingMaxId;
        return id;
    }

    /**
     * Removes a node to the object of all currently playing nodes.
     * @param {int} id The id of the node to remove.
     */
    function removeFromPlayingNodes(id) {
        delete currentlyPlayingNodes[id];
    }

    /**
     * Returns true if the playing nodes object contains the passed id.
     * @param {int} id The id to check for.
     * @returns {boolean} True if the playing nodes object contains the passed id.
     */
    function playingNodesHas(id) {
        return currentlyPlayingNodes.hasOwnProperty(id);
    }

    /**
     * Gets the node at the given id and returns it.
     * @param {int} id The id to get.
     * @returns {object} The object containing both the node and the gain, if it exists, or undefined otherwise.
     */
    function playingNodesGet(id) {
        return currentlyPlayingNodes[id];
    }

    /**
     * Starts looping the audio.
     * Always starts with a fade.
     */
    function startLoop() {
        var source = context.createBufferSource();
        source.buffer = audioBuffer;
        source.loop = true;
        source.connect(loopGain);
        loopGain.gain.setValueAtTime(0.001, context.currentTime);
        loopGain.gain.exponentialRampToValueAtTime(ownVolume, context.currentTime + that.fadeTime);
        source.start();
        loopNode = source;
        looping = true;
    }

    /**
     * Queues the audio file to start looping when loaded.
     */
    function queueLoop() {
        queuedLoop = true;
    }

    /**
     * Immediately stops an ongoing loop.
     */
    function stopOngoingLoop() {
        loopNode.stop(context.currentTime);
        looping = false;
    }

    /**
     * Cancels a queued audio file loop.
     */
    function stopQueuedLoop() {
        queuedLoop = false;
    }

    /**
     * Fades out an ongoing loop.
     */
    function fadeOngoingLoop() {
        loopGain.gain.exponentialRampToValueAtTime(0.001, context.currentTime + that.fadeTime);
        loopNode.stop(context.currentTime + that.fadeTime);
        looping = false;
    }

    /**
     * Cancels a queued audio file loop.
     */
    function fadeQueuedLoop() {
        stopQueuedLoop();
    }

    // PUBLIC MEMBERS \\        this.public = "test";

    // URL the audio file was loaded from. Readonly.
    this.url = url;

    // Time required for looping music to start and stop fading.
    this.fadeTime = 1;

    // PUBLIC FUNCTIONS \\      this.publicFunction = function(){};

    /**
     * Causes this AudioFile to start loading.
     * Does nothing if already loading or loaded or failed to load.
     */
    this.load = function () {
        if (loading || loaded || error) {
            return;
        }
        loading = true;
        createPromise();
        startRequest();
    };

    /**
     * Plays this audio file a single time.
     * Returns the ID of the playing node.
     * Does nothing if the audio file has not yet finished loading.
     * @returns {int} The id of the playing node, which can be used to stop it.
     */
    this.play = function () {
        if (!loaded) {
            return;
        }
        var volumeGain = createVolumeGain(ownVolume);
        volumeGain.connect(effectsVolumeGain);
        var source = context.createBufferSource();
        source.buffer = audioBuffer;
        source.connect(volumeGain);
        source.start();
        return addToPlayingNodes(source, volumeGain);
    };

    /**
     * Stops the playback referenced by the id, if such a playback exists.
     * @param {int} id The id of the play to stop.
     */
    this.stop = function (id) {
        if (!playingNodesHas(id)) {
            return;
        }
        var object = playingNodesGet(id);
        if (!object) {
            return;
        }
        var node = object.node;
        if (!node) {
            return;
        }
        node.stop(context.currentTime);
    };

    /**
     * Stops all currently playing sounds apart from the loop.
     */
    this.stopAll = function () {
        for (var id in currentlyPlayingNodes) {
            if (currentlyPlayingNodes.hasOwnProperty(id)) {
                that.stop(id);
            }
        }
    };

    /**
     * Fades the playback referenced by the id, if such a playback exists.
     * @param {int} id The id of the play to fade out.
     */
    this.fade = function (id) {
        if (!playingNodesHas(id)) {
            return;
        }
        var object = playingNodesGet(id);
        if (!object) {
            return;
        }
        var node = object.node;
        var gain = object.gain;
        if (!node || !gain) {
            return;
        }
        gain.gain.exponentialRampToValueAtTime(0.001, context.currentTime + that.fadeTime);
        node.stop(context.currentTime + that.fadeTime);
    };

    /**
     * Fades all currently playing sounds apart from the loop.
     */
    this.fadeAll = function () {
        for (var id in currentlyPlayingNodes) {
            if (currentlyPlayingNodes.hasOwnProperty(id)) {
                that.fade(id);
            }
        }
    };

    /**
     * Plays this audio file a single time after it has loaded, or immediately if it has already loaded.
     */
    this.playWhenLoaded = function () {
        if (loaded) {
            return that.play();
        }
        queuedPlay = true;
        return currentlyPlayingMaxId;
    };

    /**
     * Cancels playing an audio file when loaded.
     * Can only be cancelled before the file has finished loaded.
     * If the file has already loaded, this function has no effect.
     */
    this.cancelPlayWhenLoaded = function () {
        queuedPlay = false;
    };

    /**
     * Plays this audio file over and over again.
     * Does nothing if the audio file is already looping.
     * Waits for the audio to load before attempting to play it.
     */
    this.loop = function () {
        if (error || looping || queuedLoop) {
            return;
        }
        if (loaded) {
            startLoop();
        } else {
            queueLoop();
        }
    };

    /**
     * Immediately stops this looping audio file.
     * Does nothing if the audio file is not looping.
     * Prevents audio files that have not yet loaded, but have already received the instruction to loop, from looping.
     * Has no effect on audio started by the play() function.
     */
    this.stopLoop = function () {
        if (error || (!queuedLoop && !looping)) {
            return;
        }
        if (loaded) {
            stopOngoingLoop();
        } else {
            stopQueuedLoop();
        }
    };

    /**
     * Fades out this looping audio file.
     * Does nothing if the audio file is not looping.
     * Prevents audio files that have not yet loaded, but have already received the instruction to loop, from looping.
     * Has no effect on audio started by the play() function.
     */
    this.fadeLoop = function () {
        if (error || (!queuedLoop && !looping)) {
            return;
        }
        if (loaded) {
            fadeOngoingLoop();
        } else {
            fadeQueuedLoop();
        }
    };
}

